use std::collections::HashMap;

use crate::{handling::routes::Data, utils::mime::mime_enum::Mime};

use super::{datatypes::ParseErrors, ParseFormError, Request};
static TWO_NEWLINES: u8 = 3;

impl Request<'_> {
    /// # Gets data from a get_form as a HashMap
    ///
    /// # Errors
    /// Gives back a [ParseFormError], top level, if there is lacking data
    ///
    /// If everything is fine on the top level it gives back a HashMap of keys and Results, that
    /// indicate wether the key exists with the [ParseFormError] with an error of
    /// [ParseErrors::NoData] or wether the key is corrupt with the [ParseErrors::BadData]-Variant
    ///
    /// # Examples
    /// ```
    /// let request = Request {
    ///     uri: "/form?name=Name&age=Age",
    ///     headers: vec![],
    ///     method: Method::Get,
    ///     cookies: None,
    ///     mime_type: None,
    /// };
    /// let right = request.get_get_form_keys(&["name", "age"]).unwrap();
    /// assert_eq!(&"Name", right.get("name").unwrap().as_ref().unwrap());
    /// assert_eq!(&"Age", right.get("age").unwrap().as_ref().unwrap());
    ///
    /// let wrong_request = Request {
    ///     uri: "/form",
    ///     ..request.clone()
    /// };
    /// assert_eq!(
    ///     Err(ParseFormError {
    ///         error: ParseErrors::NoData
    ///     }),
    ///     wrong_request.get_get_form_keys(&["name", "age"])
    /// );
    ///
    /// let bad_data = Request {
    ///     uri: "/form?age=",
    ///     ..request.clone()
    /// };
    /// let wrong = bad_data.get_get_form_keys(&["name", "age"]).unwrap();
    /// assert_eq!(
    ///     &Err(ParseFormError {
    ///         error: ParseErrors::NoData
    ///     }),
    ///     wrong.get("name").unwrap()
    /// );
    /// assert_eq!(&Ok(""), wrong.get("age").unwrap());
    /// ```
    ///
    /// # Panics
    /// No Panics
    pub fn get_get_form_keys<'a>(
        &'a self,
        keys: &'a [&str],
    ) -> Result<HashMap<&str, Result<&str, ParseFormError>>, ParseFormError> {
        let data = if let Some(val) = self.uri.split_once('?') {
            val
        } else {
            return Err(ParseFormError {
                error: ParseErrors::NoData,
            });
        };
        let data = data
            .1
            .split('&')
            .map(|kvp| kvp.split_once('='))
            .collect::<Vec<Option<(&str, &str)>>>();

        let mut values: HashMap<&str, &str> = HashMap::new();
        for kvp in data {
            let kvp = if let Some(kvp) = kvp {
                kvp
            } else {
                continue;
            };
            values.insert(kvp.0, kvp.1);
        }
        let mut response = HashMap::new();
        for key in keys {
            let entry = if let Some(val) = values.get(key) {
                Ok(*val)
            } else {
                Err(ParseFormError {
                    error: ParseErrors::NoData,
                })
            };
            response.insert(*key, entry);
        }
        Ok(response)
    }
    pub fn get_post_data<'a>(
        &'a self,
        keys: &[&'a str],
        data: &Data,
    ) -> Result<HashMap<String, Result<Vec<u8>, ParseFormError>>, ParseFormError> {
        let data = data.buffer.as_slice();
        let mut keymap: HashMap<String, Result<Vec<u8>, ParseFormError>> =
            HashMap::with_capacity(keys.len());

        for key in keys {
            keymap.entry(key.to_string()).or_insert(Err(ParseFormError {
                error: ParseErrors::NoData,
            }));
        }
        let Some(ref mime_type) = self.mime_type else {
            return Err(ParseFormError { error: ParseErrors::BadData });
        };

        match mime_type.0 {
            Mime::ApplicationXWwwFormUrlencoded => {
                let Ok(data) = String::from_utf8(data.to_vec()) else {
                    return Err(ParseFormError { error: ParseErrors::BadData });
                };
                for kvp in data.split('&') {
                    let Some(mut kvp) = kvp.split_once('=') else {
                        return Err(ParseFormError { error: ParseErrors::BadData });
                    };

                    let Some(thing) = keymap.get_mut(kvp.0) else {
                        continue;
                    };
                    kvp.1 = kvp.1.trim_end_matches('\0');

                    *thing = Ok(kvp.1.as_bytes().to_vec());
                }
            }
            Mime::MultipartFormData => {
                let Some(mut boundary) = mime_type.1.split(';').find(|element| element.trim().starts_with("boundary=")) else {
                    return Err(ParseFormError { error: ParseErrors::BadData });
                };
                boundary = boundary
                    .trim()
                    .strip_prefix("boundary=")
                    .unwrap()
                    .trim_matches('"');
                let mut temp_bound = "--".to_string();
                temp_bound.push_str(boundary);
                let end_boundary = format!("{temp_bound}--\r").as_bytes().to_owned();
                temp_bound.push('\r');
                let boundary = temp_bound.as_bytes();
                Self::get_multipart_data(data, boundary, &end_boundary, &mut keymap);
            }
            _ => {
                return Err(ParseFormError {
                    error: ParseErrors::BadData,
                })
            }
        };
        Ok(keymap)
    }
    fn get_multipart_data(
        data: &[u8],
        boundary: &[u8],
        end_boundary: &[u8],
        map: &mut HashMap<String, Result<Vec<u8>, ParseFormError>>,
    ) {
        let mut current_part: Vec<&[u8]> = vec![];
        let mut current_key: Option<String> = None;
        let mut ignore_line = 0;
        for part in data.split(|byte| byte == &b'\n') {
            if part == [b'\r'] {
                if current_key.is_some() {
                    if ignore_line >= TWO_NEWLINES {
                        current_part.push(&[b'\n']);
                        continue;
                    }
                    ignore_line += 1;
                }
                continue;
            }
            if part == end_boundary {
                if let Some(key) = &current_key {
                    let mut part = current_part.join(&b'\n');
                    if part.ends_with(&[b'\r']) {
                        part.pop();
                    }
                    map.insert(key.to_string(), Ok(part));
                }
                break;
            }
            if part == boundary {
                if let Some(key) = &current_key {
                    let mut part = current_part.join(&b'\n');
                    if part.ends_with(&[b'\r']) {
                        part.pop();
                    }
                    map.insert(key.to_string(), Ok(part));
                }
                current_part = vec![];
                current_key = None;
                ignore_line = 0;
                continue;
            }
            if part.starts_with(b"Content-Disposition: form-data; name=") {
                let headers = part
                    .split(|byte| byte == &b';')
                    .filter(|header| !header.is_empty())
                    .collect::<Vec<_>>();
                if headers.len() < 2 {
                    continue;
                }
                let name = headers[1].split(|byte| byte == &b'=').collect::<Vec<_>>();
                if name.len() != 2 {
                    continue;
                }
                let mkey = String::from_utf8_lossy(name[1])
                    .as_ref()
                    .trim_end()
                    .trim_matches('"')
                    .to_owned();

                if map.contains_key::<str>(&mkey) {
                    current_key = Some(mkey.to_owned());
                }
                continue;
            } else if current_key.is_some() {
                current_part.push(part);
            }
        }
    }
}
#[cfg(test)]
mod test {
    use crate::{
        handling::{
            methods::Method,
            request::{datatypes::ParseErrors, ParseFormError},
            routes::Data,
        },
        utils::mime::mime_enum::Mime::{ApplicationXWwwFormUrlencoded, MultipartFormData},
    };

    use super::Request;
    #[test]
    fn try_get_test() {
        let request = Request {
            uri: "/form?name=Name&age=Age",
            headers: vec![],
            method: Method::Get,
            cookies: None,
            mime_type: None,
        };
        let right = request.get_get_form_keys(&["name", "age"]).unwrap();
        assert_eq!(&"Name", right.get("name").unwrap().as_ref().unwrap());
        assert_eq!(&"Age", right.get("age").unwrap().as_ref().unwrap());

        let wrong_request = Request {
            uri: "/form",
            ..request.clone()
        };
        assert_eq!(
            Err(ParseFormError {
                error: ParseErrors::NoData
            }),
            wrong_request.get_get_form_keys(&["name", "age"])
        );

        let bad_data = Request {
            uri: "/form?age=",
            ..request.clone()
        };
        let wrong = bad_data.get_get_form_keys(&["name", "age"]).unwrap();
        assert_eq!(
            &Err(ParseFormError {
                error: ParseErrors::NoData
            }),
            wrong.get("name").unwrap()
        );
        assert_eq!(&Ok(""), wrong.get("age").unwrap());
    }

    #[test]
    fn try_post_text() {
        let req = Request {
            uri: "",
            headers: vec!["Content-Type: application/x-www-form-urlencoded".to_string()],
            method: Method::Post,
            cookies: None,
            mime_type: Some((ApplicationXWwwFormUrlencoded, "".into())),
        };
        let data = Data {
            buffer: b"message=23&message1=24".to_vec(),
            is_complete: true,
        };
        let map = req.get_post_data(&["message", "message1"], &data).unwrap();
        assert_eq!(
            &b"23".to_vec(),
            map.get("message").unwrap().as_ref().unwrap()
        );
        assert_eq!(
            &b"24".to_vec(),
            map.get("message1").unwrap().as_ref().unwrap()
        );
        let req = Request {
            uri: "",
            headers: vec!["Content-Type: multipart/form-data; boundary=\"boundary\"".to_string()],
            method: Method::Post,
            cookies: None,
            mime_type: Some((
                MultipartFormData,
                "charset=UTF-8; boundary=\"boundary\"".into(),
            )),
        };
        let data = Data {
            buffer: b"--boundary\r
Content-Disposition: form-data; name=\"field1\"\r
\r
value1\r
--boundary\r
Content-Disposition: form-data; name=\"field2\"; filename=\"example.txt\"\r
\r
va\nlue2\r
--boundary--\r
"
            .to_vec(),
            is_complete: true,
        };
        let map = req.get_post_data(&["field1", "field2"], &data).unwrap();

        assert_eq!(
            &b"value1".to_vec(),
            map.get("field1").unwrap().as_ref().unwrap()
        );
        assert_eq!(
            &b"va\nlue2".to_vec(),
            map.get("field2").unwrap().as_ref().unwrap()
        );
    }
}
